%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        ML-CAM Translation                             %%%
%%%        This version uses de Bruijn indexing to        %%%
%%%        generate code for identifiers.                 %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

module camml.

import mldecl camdecl lists.


type  translate        tm -> prog -> o.
type  trans            (list (pair tm int)) -> int -> tm -> com -> o.
type  transop          tm -> oper -> o.
type  generate_path    int -> com -> o.
type  member           A -> (list A) -> o.



%%% (translate ML CAM) translates mini-ML expression `ML' into CAM program
%%% `CAM'.

translate E (program C) :-
        trans [] 0 E C.



trans Gamma Depth (# N) (quote (## N)) :- !.     %%% (1) Numbers

trans Gamma Depth tt (quote ctt) :- !.           %%% (2) true
trans Gamma Depth ff (quote cff) :- !.           %%% (3) false
trans Gamma Depth nil (quote cnil) :- !.

trans Gamma Depth (cond E1 E2 E3)                %%% (5) Conditional
      (push & C1 & (branch C2 C3)) :- !,
        trans Gamma Depth E1 C1,
        trans Gamma Depth E2 C2,
        trans Gamma Depth E3 C3.

trans Gamma Depth (E1 && E2)                     %%% (6) Product
      (push & C1 & swap & C2 & cons) :- !,
        trans Gamma Depth E1 C1,
        trans Gamma Depth E2 C2.

trans Gamma Depth (Op @ E) (push & C & (op Camop)) :- %%% (7,8) Projections
        member Op [fst, snd], !,
        trans Gamma Depth E C,
        transop Op Camop.

trans Gamma Depth  (lamb E)  (cur C) :- !,       %%% (9) Lamb
        NewDepth is Depth + 1,
        pi K\(trans [(pair K NewDepth)|Gamma]
                    NewDepth  (E K) C).

trans Gamma Depth (let M E2)                     %%% (11) Let
      (push & C2 & cons & C1) :- !,
        trans Gamma Depth E2 C2,
        NewDepth is Depth + 1,
        pi K\(trans [(pair K NewDepth)|Gamma]  
                    NewDepth  (M K) C1).

trans Gamma Depth (fix M) (push & (rec C)) :- !, %%% (12) Fix
        NewDepth is Depth + 1,
        pi K\(trans [(pair K NewDepth)|Gamma]
                    NewDepth  (M K) C).

                                                 %%% (10) Application
trans Gamma Depth (Op @ E1 @ E2)                 %%% binary primitives
      (push & C1 & swap & C2 & cons & (op Camop)) :-
        member Op [==, !=, lt, gt, ::, plus, minus, times], !,
        trans Gamma Depth E1 C1,
        trans Gamma Depth E2 C2,
        transop Op Camop.

trans Gamma Depth (Op @ E1)                      %%% unary primitives
      (C1 & (op Camop)) :-
%      (push & C1 & (op Camop)) :-
        member Op [hd, tl, empty], !,
        trans Gamma Depth E1 C1,
        transop Op Camop.

trans Gamma Depth (E1 @ E2)                      %%% General Apply
      (push & C1 & swap & C2 & cons & app) :- !,
        trans Gamma Depth E1 C1,
        trans Gamma Depth E2 C2.

trans Gamma Depth K C :-                         %%% (4) Identifier 
        member (pair K D) Gamma,
        Distance is (Depth - D) + 1,
        generate_path Distance C.


%%% translations for primitive operators %%%
transop ==    cequal.
transop !=    cnequal.
transop lt    clt.
transop gt    cgt.
transop ::    ccons.
transop plus  cplus.
transop minus cminus.
transop times ctimes.
transop hd    chd.
transop tl    ctl.
transop empty cempty.


%%% translate N+1 into (cfst^N & csnd) %%%
generate_path 1 csnd :- !.
generate_path N (cfst & C) :- !,
        N1 is N - 1,
        generate_path N1 C.

%%%   End of module camml   %%%

