
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                         %
%                           FILE  SCANNER.PL                              %
%                         --------------------                            %
%                                                                         %
%    This file contains a set of clauses that implement a scanner used    %
%   principally to read text (logical) line by line from a file. The main %
%   predicates in this regard are read_line(Line) and read_in(Line,Eoln). %
%   This code is adapted from Clocksin & Mellish and from Dale's imp:lex. %
%   The major difference from C&M is that an entire file has to be read.  %
%   The differences from Dale's code are that (1) the entire file is not  %
%   read in one shot, but rather text is read line by line, and (2) no    %
%   end-of-file marker is used but rather this is discovered by detecting %
%   an error condition. A lookahead character is used while determining   %
%   tokens. Consequently, the first character of the next line is read in %
%   the course of reading the current line. In order to make line by line %
%   processing possible, a trace of this character must be left, and this %
%   is done via an assertion of the form look_ahead(Char). The important  %
%   thing to note is that prior to reading a file, this assertion needs   %
%   to be initialised (eg look_ahead(32) - 32 is ASCII for blank - could  %
%   be used. The second point to note is that the detection of an end-    %
%   of-file condition depends crucially on the attempt to read beyond the %
%   end of a file being noted merely as a failure, and not being reported %
%   as an error. In other words, before a file is opened to be read, the  %
%   nofileerrors option must be turned on in CProlog. Finally, if an end- %
%   of-file is encountered then this results in an empty line being       %
%   returned by read_line. Any external routines that use this scanner    %
%   should check for this in order to determine whether the file is read  %
%   fully.                                                                %
%                                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


    :- dynamic look_ahead/1.

%          initialising some `constant' definitions

% the ASCIIs for characters that have a special role
    comment(0'%).       % start of a comment
    quote(0'').         % start of a string
    quote(0'").         % Dale used this for strings. Why?

%  The next three lines are needed for  Quintus version 1.6
%   is_space(Char) :-   ( name(' ',[Char]) | name('	',[Char])), !.
%   is_endfile(-1).      
%   is_newline(10).     

    semicolon(';').

% Flush the rest of the (physical) line in the input text
     skip_line :- 
         repeat,
             get0(C),
             ( is_endfile(C) -> !, fail
             | is_newline(C) -> true
 	    | fail ), !.

% Note: The above is Fernando's improvement of
%     skip_line :- skip(10).

%  Get the next char in the input, skipping over spaces, comments, and
%  end-of-line; Char is the lookahead. Fails if end-of-file is encountered
    get_next_non_space(Char,NChar) :-
     ((is_space(Char),! ; comment(Char), !, skip_line ; is_newline(Char),!),
          get0(Char1), get_next_non_space(Char1,NChar) ;
       is_endfile(Char), !, fail ;
       NChar = Char).

%  Get the first character of the current line
  get_lookahead(Char) :- (retract(look_ahead(Char)) ; is_space(Char)), !.

%  Store the first character of the next line
  put_lookahead(Char) :- assert(look_ahead(Char)).

%  Read the next (logical) line in the input and return it as a list
%  of tokens.
     read_line(Line) :- read_in(Line,'.'), !.


%  Read the sequence of tokens upto the first one equal to EndWord and return
%  it as a list. If an end-of-file is encountered before EndWord then return
%  an empty list; in this case if the sequence is nonempty then flag an error.
  read_in(Words,EndWord) :-
      get_lookahead(Char), read_word(Char,Word,NextChar), !,
      (rest_sent(Word,NextChar,FinalChar,RestWords,EndWord), !,
                 Words = [Word|RestWords], put_lookahead(FinalChar) ;
       read_error(eof,EndWord), Words = []).

  read_in([],_).



%  Read in the next word from the input. Fail if end-of-file is encountered
%  before the start of a word. Char and NextChar are the lookaheads on entry
%  and on exit respectively
  read_word(Char,Word,NextChar) :- 
      get_next_non_space(Char,TChar),
       (quote(TChar), !, read_string(TChar,Word,NextChar) ;
        read_word1(TChar,Word,NextChar),! ; 
        read_error(badchar,TChar), 
             is_space(NChar), read_word(NChar,Word,NextChar)).


 read_word1(Char,Word,NextChar) :-
     single_character(Char), !, name(Word,[Char]), get0(NextChar).

 read_word1(Char,Word,NextChar) :-
     in_word(Char), !, get0(TChar),rest_word(TChar,Cs,NextChar),
     name(Word,[Char|Cs]).

% Get the remaining characters in a legal word
 rest_word(Char,[Char|Cs],NextChar) :-
     in_word(Char), !, get0(TChar), rest_word(TChar,Cs,NextChar).

 rest_word(Char,[],Char).



 read_string(Char,Word,NextChar) :-
     get0(TChar), rest_string(TChar,Str,NextChar), name(Word,[Char|Str]).


% Get the remaining characters in a legal string.
 rest_string(Char,[0''],Char) :-
     is_endfile(Char), !, read_error(eof,'''').

 rest_string(Char,String,NextChar) :-
     quote(Char),!,get0(NChr1),
     (quote(NChr1), 
        !, get0(NChr2), rest_string(NChr2,RString,NextChar),
        String = [Char|RString] ;
      NextChar = NChr1, String = [Char]) .

 rest_string(Char,[Char|String],NextChar) :- !,
     get0(Chr), rest_string(Chr,String,NextChar).


%  Get the remaining tokens upto EndWord; fail if an end-of-file is encountered
%  before an EndWord. Char and FinalChar are the lookahead characters before
%  entry and after exit respectively
  rest_sent(Word,Char,Char,[],Word) :- !.
 
  rest_sent(Word,Char,FinalChar,[NWord|Ws],EndWord) :-
      read_word(Char,NWord,NChar), rest_sent(NWord,NChar,FinalChar,Ws,EndWord).





% This is a list of characters that are legal in a word
in_word(33).           %  !  
in_word(35).           %  #
in_word(38).           %  &  
in_word(45).           %  -
in_word(47).           %  /  
in_word(58).           %  :
in_word(60).           %  <  
in_word(61).           %  =  
in_word(62).           %  >  
in_word(63).           %  ?
in_word(64).           %  @
in_word(95).           %  _
in_word(C) :- C>96, C<123.  % a,b,...,z
in_word(C) :- C>64, C<91.   % A,B,...,Z
in_word(C) :- C>47, C<58.   % 0,1,...,9





% These are a list of characters that form words by themselves
single_character(36).  %  $  
single_character(40).  %  (  
single_character(41).  %  )  
single_character(42).  %  *  
single_character(43).  %  +  
single_character(44).  %  ,  
single_character(46).  %  .  
%single_character(58).  %  :
single_character(59).  %  ;  
single_character(91).  %  [
single_character(92).  %  \  
single_character(93).  %  ]
single_character(94).  %  ^
single_character(96).  %  `
single_character(123). %  {  
single_character(124). %  |
single_character(125). %  }  
single_character(126). %  ~  



% This is sometimes useful for interacting with the user
   get_response(Word) :- is_space(Char),!, read_word(Char,Word,_).


% This is useful in removing the double quotes from around a string
   remove_quotes(Str,Sans) :- 
     name(Str,[Quote|Rest]), reverse_cdr(Rest,Sanslist), name(Sans,Sanslist).

