%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                           FILE    ERRORS.PRO                                %
%                          --------------------                               %
%                                                                             %
%   The clauses in this file are useful for providing error messages when     %
%   errors occur either in reading or in parsing various declarations. The    %
%   first argument in most cases indicates the place - i.e. the declaration - %
%   where the error routine is invoked. The clauses in this file use certain  %
%   printing routines defined in display.pro. These are the following:        %
%                                                                             %
%                                                                             %
%   write_tab          :    Expected to provide indentation on new line.      %
%                                                                             %
%   write_line(Line)   :    Display the list of tokens in Line.               %
%                                                                             %
%   newl               :    Like nl, except it provides an indentation.       %
%                                                                             %
%   write_term         :    Displays a lambda Prolog term (in untyped         %
%        (Term,Mod)    :    format).                                          %
%                                                                             %
%   write_type(Type)   :    Displays a type expression.                       %
%                                                                             %
%   write_tyterm       :    Displays a lambda Prolog term (in untyped         %
%         (Term,Ty,Mod):    format) and its type.                             %
%                                                                             %
%                                                                             %
%   The following routine defined in parser.pro is also used:                 %
%                                                                             %
%    opname(Op,Opname)   :    Given an operator term, gets the operator name. %
%                                                                             %
%                                                                             %
%   After an error message is displayed, the cursor is moved to a new line,   %
%   but no indentation is provided.                                           %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





%       Signalling an error that occurs when attempting to read from a file

% If an end of file is encountered before the line being read is over
    read_error(eof,EndWord) :-
         newl, newl, 
         write('Read Error : Attempt to read beyond end of file looking for '),
         write(EndWord), nl.

% If an unrecognised character occurs while reading a line
    read_error(badchar,Char) :-
         newl, newl, write('Read Error : Illegal character '), 
         name(Name,[Char]), write(Name),  write(' in the input stream'), nl.








%    Signalling an error that occurs when attempting to parse a declaration

% A header message that is used in all the error routines
   write_error_header(Line) :-
       write_tab, write_line(Line), nl, write_tab, write(' Syntax error : ').


% When an expected delimiter is missing
   syntax_error(delimiter,EOT,Line) :-
       write_error_header(Line),
       write(' Missing delimiter. Expecting a '), write(EOT), nl.



% If the name of an operator, kind or type declaration is not a legal token
   syntax_error(DecKind,name,Line) :-
       write_error_header(Line), write('Name of '), 
       (DecKind = kinddec, write('kind ') ;
        DecKind = typedec, write('type ') ;
        DecKind = opdec, write('operator ')),
       write('being defined must be a token'), nl.





% Errors in the module name declaration

% If the module name is not declared at all
   syntax_error(moddec,name,none,Line) :-
        !, write_error_header(Line),
        write('expecting a module name declaration'), nl.


% If the name of the module is not the expected one
   syntax_error(moddec,name,Mod,Line) :-
        write_error_header(Line),
        write('Assuming module name is '), write(Mod), nl.





% Errors in the import declaration

% If import list is not a list of tokens
   syntax_error(impdec,implist,Line) :-
        write_error_header(Line),
        write('Expecting a list of tokens in an import declaration'),nl.

   syntax_error(impdec,cycle,Mod,ImpMod,Line) :-
        write_error_header(Line),
	write(' Cycle in import heirarchy.'), newl,
	write('    Module '), write(ImpMod), 
	write(' already inherits module '), write(Mod), nl.



% Errors in the operator declaration

% If the priority specification is not a valid one
   syntax_error(opdec,priority,Line) :-
        write_error_header(Line),
        write('Priority must be an integer between 1 and 255'), nl.

% If the mode specification is incorrect or does not match the operator type
   syntax_error(opdec,mode,Fx,Line) :-
        write_error_header(Line),
        write('Mode for '), write(Fx), write(' operator must be '),
        (Fx = prefix, write('fx or fy') ;
         Fx = postfix, write('xf or yf') ;
         Fx= infix, write('xfx, xfy, yfx, or yfy')), nl. 







% Errors in a kind declaration

% An illegal symbol in a kind expression; currently only -> and TYPE are valid.
   syntax_error(kinddec,kind,token,X,Line) :-
        write_error_header(Line),
        write('Illegal symbol '), write(X), write(' in kind expression'), nl.

% If a non -> symbol follows another such symbol; currently kind constructors 
% are not permitted.
   syntax_error(kinddec,kind,appl,Sy2,Sy1,Line) :-
        write_error_header(Line),
        write('Kind constructors are not permitted'), nl.
%       write('Only a -> can follow '), write_type(Sy1),
%       write('. Got '), write(Sy2), write(' instead.'), nl.
%   Problem: internal representation for Sy1 may not correspond to the input






% Some errors that are common between kind and type declarations

% If the kind or type expression is missing in a kind or type definition
   syntax_error(DecKind,Field,none,Line) :-
        write_error_header(Line),
        write('Missing '), write(Field), 
        write(' expression in '), write(Field), write(' declaration'), nl.

% if the domain or range kind (type) in a functional kind (type) is missing
   syntax_error(DecKind,KorT,fn,RorD,none,Line) :-
        write_error_header(Line),
        write('Missing '), write(RorD), write(' '),
        write(KorT), write(' for a functional in '), 
        write(KorT), write(' declaration'),nl.





% Errors in the type declaration

% If a symbol in a type expression has no defined kind
   syntax_error(typedec,type,symbolkind,Sy,Line) :-
       write_error_header(Line),
       write('Symbol '), write(Sy), 
       write(' in type expression has no kind'), nl.

% The kind of the range or domain in a functional type must be TYPE 
   syntax_error(typedec,type,fn,DorR,kind,X,XK,Line) :-
       write_error_header(Line),
       write('Kind of '), write(DorR), 
       write(' in a function type must be TYPE'), nl.

% if the kind of the type expression in a type declaration is not TYPE
   syntax_error(typedec,type,kind,Ty,Kind,Line) :-
	write_error_header(Line),
	write('Expecting a type expression of kind TYPE. Got'), newl,
	write('    '), write_type(Ty), 
	write('  of kind  '), write_type(Kind), write('  instead.'), nl.

% if a 0-ary type constructor is given an argument
   syntax_error(typedec,type,appl,[R1,R1K],[R2,R2K],Line) :-
       write_error_header(Line),
       write('Type expression of kind  '), write_type(R1K),
       write('  cannot be applied to'), newl,
       write('                expression of kind '), write_type(R2K),nl.







% Errors in the clause or expression declarations

% if object being defined (i.e. function binding) is not a token
   syntax_error(cande_dec,exp,tokenname,Line) :-
     write_error_header(Line),
     write('Only a non-priimitive constant can be bound by a definition'), nl.


% if the types are incompatible in an attempted binding
   syntax_error(cande_dec,exp,binding,Tok,Ty,Term,Ty1,Line) :-
     write_error_header(Line),
     write('Types are incompatible. Cannot bind'),newl,
     write('   '), write_tyterm(Tok,Ty,Mod), write(' to'), newl,
     write('   '), write_tyterm(Term,Ty1,Mod), nl.


% if head of a clause does not have a nonlogical constant head
   syntax_error(cande_dec,clause,head,Line) :-
     write_error_header(Line),
     write(' Name of predicate defined by a clause must be a constant.'), nl.


% If a term expected to be a clause is actually not one
   syntax_error(cande_dec,clause,notclause,Ty,Line) :-
     write_error_header(Line),write('Expecting a clause. Got a term of type '),
     write_type(Ty), write(' instead.'), nl.








% Errors in a lambda Prolog term

% If the term is missing
   syntax_error(term,none,Line) :-
         write_error_header(Line), 
         write('Expecting a term. '), nl.



% If an operator is not of the right kind (i.e. prefix, infix, or postfix)
   syntax_error(term,oper_kind,Fx,Op,Line) :-
     write_error_header(Line), write('Operator '),
     (Fx = prefixfollow, write('immediately following ') ; true),
     opname(Op,Opname), write(Opname),
     write(' must be prefix'), nl.

% If the argument to an operator is missing
   syntax_error(term,oper_arg,Op,Line) :-
     write_error_header(Line),
     write('Operator '), opname(Op,Opname), write(Op),
     write(' Expects an argument'),nl.

% If there is an error in operator precedence
   syntax_error(term,oper_prec,LGE,Op1,Op2,Line) :-
     write_error_header(Line),
     write('Incorrect operator precedence.'), newl, 
     write('               '), opname(Op2,Opname2), write(Opname2),
     (LGE = lower, write(' must have lower right precedence than ');
      LGE = higher, write(' must have higher left precedence than ');
      LGE = neq,
       write(' must have left precedence unequal to right precedence of ')),
     opname(Op1,Opname1), write(Opname1), nl.

% If there is an attempt to abstract a non variable term
   syntax_error(term,abst,Term,Mod,Line) :-
     write_error_header(Line),
     write(' Only variables can be abstracted'), newl,
     write('       Cannot abstract '), write_term(Term,Mod), nl.

% If types are incompatible in an attempted application
syntax_error(term,appl,Tortyp,Torterm,Candtyp,Candterm,Mod,Line) :-
  write_error_header(Line),
  write(' Types are incompatible. '),newl,
  write('    Cannot apply '), write_tyterm(Torterm,Tortyp,Mod), newl,
  write('    to           '), write_tyterm(Candterm,Candtyp,Mod), nl.


% If the type provided by the user within a term cannot be resolved with 
% the other constraints.
syntax_error(term,type,CorV,UsrType,ConstrType,Line) :-
  write_error_header(Line),
  write('The term '), write(CorV), 
  write(' must have a type of the form '), newl,
  write('     '), write_type(ConstrType), newl,
  write('It cannot have the type '), write_type(UsrType), nl.


syntax_error(term,list,oper,Op,Line) :-
  write_error_header(Line),
  Op = '|', write('Right operand of | must be a list.'),
  !, nl.

syntax_error(term,list,oper_arg,Op,Line) :-
  (not (Op = ',' ; Op = '|'), syntax_error(term,list,oper,Op,Line), ! ;
   write_error_header(Line),
   write('Missing tail in list expression.'), nl).

% The type of a goal must be boolean
   syntax_error(goal,type,Type,Line) :-
       write_error_header(Line), 
       write('Expecting a goal; got a term of type '), write_type(Type),
       write('.'), nl.





% Some runtime errors. These need to be worked on soon.
   runtime_error(trace, var) :-  
           writeblanks(2), 
           write('Error in predicate: Tracing option must not be a variable'),
           nl, !.

   runtime_error(trace, option, Opt) :-  
           writeblanks(2), 
           write('Invalid tracing option. Valid options are'),
	   nl, writeblanks(3), 
	   write('clause, goal, interp, unifier, hyps, and allopts'), nl, !.

