
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                             %
%                               FILE  INT.PRO                                 %
%                               -------------                                 %
%                                                                             %
%                                                                             %
%  This file contains the top-level of the lambda Prolog logic programming    %
%  interpreter.                                                               %
%                                                                             %
%  This version of int.pro contains some simple code to deal with             %
%  universally quantified goals.  The handling of universally quantified      %
%  goal is not described elsewhere.  A report on how it is managed will be    %
%  forthcoming.  The main feature of its implementation here is that the      %
%  unification algorithm of Huet's (implemented in unify.pro) does not        %
%  need to be modified.  The kind of disagreement pairs given to the unifer,  %
%  however, is a bit more complex than if universal goals were not permitted. %
%                                                                             %
%  This interpreter implements the following goal forms:                      %
%     (1) disjunctions,                                                       %
%     (2) conjunctions,                                                       %
%     (3) existentials,                                                       %
%     (4) universal,                                                          %
%     (5) special goals, implementing special functionality.  These are       %
%             actually implemented in the file specials.pro.                  %
%     (6) cut and negation-by-failure.  The way these two meta-logical        %
%             features interact with universal goals is not well understood   %
%	     yet.  Their interaction with the other goal forms seems          %
%	     quite reasonable, however.                                       %
%     (7) Atomic goals (of course).  Backchaining on programs is performed    %
%             with these goals.                                               %
%                                                                             %
%  Notice that implications are not permitted in goals.  Thus, full           %
%  hereditary Harrop formulas are not actually implemented (see paper in      %
%  Logic in Computer Science conference).  Implementing this feature would    %
%  have changed our current implemention too much.  In a new implementation   %
%  of lambda Prolog we expect to include this feature.                        %
%                                                                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



% The READ-EVAL loop for the lambda Prolog interpreter; The loop
% is infinite, and that termination occurs via a special goal
lp :- 
    setup_env(sys_defs),
    repeat, read_goal(main,Goal),
            freezetypevars(Goal), 
	    retractall(current_num(_,_)),   % clean up the gensym counters
            lnorm(Goal,Gnorm),              % lambda-normalize the goal
	    get_imports(main,Modules),
            interpret_goal(Modules,Gnorm),
    fail.


  freezetypevars(Goal) :- freezetypes, freeze_typevar_in_term(Goal), !.
  freezetypevars(_).


% Loading in a set of system definitions. These are assumed to be in a module
% called sys_defs, which is inherited by every other module. The inheritance
% is enforced by virtue of the initialisation to mlink in parser.pro and the
% definition of special_goal in the specials file. The system definitions are 
% essential to the functioning of the system, and hence an attempt to invoke 
% lp fails if the module sys_defs cannot be found.

   setup_env(Mname) :- 
          module_list(L), member(Mname,L),!,nl.

   setup_env(Mname) :-
          (make_file_name(Mname,'.mod',Fname), !, sign_on, nl,
               nl, write('Initialising system definitions. Loading module '),
               write_mod_name(Mname),nl,inctab, open_file(OFile,Fname), 
               load_modules([Mname]), close_file(OFile), dectab, nl ;
           write('Cannot load LP: Module '), write_mod_name(Mname),
	   write(' is missing'), nl, nl, fail).



% Reading a goal in the context of a module
   read_goal(Module,Goal) :-
       repeat, 
            write('?- '), read_line(Line),
            parse_term(Module,'.',TTerm,[],[],[],_,[],_,[],_,Line,Err,Line,[]),
        Err = false,
        (TTerm = [Goal,o], ! ;
         TTerm = [_,Ty], syntax_error(goal,type,Ty,Line), fail).



% Interpreting a goal. An attempt is made to prove the goal, and an answer
% is reported. Alternative solutions may be requested. The last cut prevents
% retrying a previous goal because of failure in the lp loop

   interpret_goal(Modules,Goal) :-
      free_variables(Goal,Ansvar), 
      prove_goal(Modules,Goal,[],Ans,FFList,[]),
      report_ans(Ansvar,Ans,FFList), !.

   interpret_goal(_,_) :- nl, write(' no'), nl, !.


   report_ans(Varlist,Subs,FFlist) :-
	( Varlist = [], !  ; 
	  report_subs(Varlist,Subs,FFlist)
	), write(' yes'), nl, !.


   report_subs(Varlist,Subs,FFlist) :-
	restrict_subs(Varlist,Subs,RelSubs), 
	write_subs(RelSubs,main),
	(FFlist = [], !, try_again;
	 nl, nl, 
	   write(' The remaining flexible - flexible pairs:'),
	   write_node(FFlist,main),
	   !, ff_options(Varlist,FFlist,Subs)
	).

   try_again :-
	  write(' '), get_response(Word),
          (semicolon(Word), !, fail ; true).


   ff_options(Varlist,FFs,Subs) :-
	write(' '), get_response(Word),
	(semicolon(Word), !, fail ;
	 clean_ff_option(Word), !, 
		answer_with_cleaned_ff(Varlist,FFs,Subs) ; 
	 triv_ff_option(Word), !, 
		answer_with_triv_ff(Varlist,FFs,Subs) ;
	 true), !.

   clean_ff_option('c').
   triv_ff_option('t').

   answer_with_triv_ff(Varlist,FFs,Subs) :-
	triv_ff_subs(FFs,Cs), compose_subs(Subs,Cs,NSubs),
	( nl, nl, 
		write(' The answer with the trivial solution to flex-flex pairs:'), 
		nl, report_subs(Varlist,NSubs,[]) ;
	  nl, nl, write(' Backing up over the trivial solution:'), nl,
		report_subs(Varlist,Subs,FFs)), !.
	  

   answer_with_cleaned_ff(Varlist,FFs,Subs) :-
	clean_ff_pairs(FFs,NFFs,Cs), 
	compose_subs(Subs,Cs,NSubs),
	nl, nl, write(' The answer after cleaning the flex-flex pairs:'),
	report_subs(Varlist,NSubs,NFFs).

   clause_instan(Uvars,Goal,Clause,Modules,NewModules) :- 
	term_head(Goal,Pred), 
	clause_instance(Modules,NewModules,_,Pred,Cl),
	para_vars_for_free_vars(Uvars,Cl,Clause).

    clause_instance(Modules,NewModules,Source,Pred,Clause) :-
	memb(Source,Modules), 
	get_imports(Source,Imports),
        join_at_end(Imports,Modules,NewModules),
        lp_clause(Source,Pred,Clause).


%  prove_goal(Modules,Goal,Dp,Cs,Succ,Uvars) is true if Goal and the set
%  of (flex-flex) disagreement pairs can be reduced to just the set of
%  (flex-flex) disagreement pairs Succ. If so Cs is the compound substitution
%  that enables this reduction. Uvars is the list of universal variables in
%  force.  The indirection - i.e. the use of prove_body is necessary
%  to deal with CUT. 

  prove_goal(Modules,Goal,Dp,Cs,Succ,Uvars) :-
        prove_body(Modules,Goal,Dp,Cs,Succ,Backtrk,Uvars),
        (Backtrk = yes ; Backtrk = no, !, fail).



%  prove_body is like prove_goal, except it has two success states. When
%  Backtrk is instantiated to yes, this is a genuine success. When it is
%  instantiated to no, this is a failure, but the failure is caused by an
%  attempt to backtrack over a CUT. A failure of this sort needs to be 
%  indicated as a success, so that an attempt is not made to try the second
%  alternative in a disjunctive goal.

 prove_body(Modules,cv(!,o,c),Dp,[],Dp,Backtrk,Uvars) :-
                             (Backtrk = yes ; !, Backtrk = no).

 prove_body(Modules,cv(;,o -> o -> o,c) ^ Goal1 ^ Goal2,Dp,Cs,Succ,Backtrk,Uvars) :-
      !, (prove_body(Modules,Goal1,Dp,Cs,Succ,Newbktrk,Uvars), 
              (Newbktrk = yes, Backtrk = yes; Newbktrk = no, !, Backtrk = no);
          prove_body(Modules,Goal2,Dp,Cs,Succ,Newbktrk,Uvars), 
              (Newbktrk = yes, Backtrk = yes; Newbktrk = no, !, Backtrk = no)).

 prove_body(Modules,cv(',',o -> o -> o,c) ^ Goal1 ^ Goal2,Dp,Cs,Succ,Backtrk,Uvars) :-
          !, prove_body(Modules,Goal1,Dp,Cs1,Succ1,Backtrk1,Uvars),
             (Backtrk1 = yes, apply_subs_to_goal(Cs1,Uvars,Goal2,NewUvars,Newgoal2),
                         prove_body(Modules,Newgoal2,Succ1,Cs2,Succ,Backtrk,NewUvars),
                         append(Cs1,Cs2,Cs);
              Backtrk1 = no, !, Backtrk = no).

% Negation-by-failure.  No substitutions are passed back and the set of
% flex-flex pairs is the same. Who knows if this is sensible, especially
% with the universal variables in there?
 prove_body(Modules,cv(not,o -> o,c)^Goal,Dp,[],Dp,yes,Uvars) :- 
	  !, not (prove_goal(Modules,Goal,Dp,_,_,Uvars)).

 prove_body(Modules,cv(pi,(A -> o) -> o,c)^B,Dp,Cs,Succ,yes,Uvars) :- 
	  !,
          new_uni_variable(cv(_,A,v),Var), lnorm(B^Var,Goal),
	  prove_goal(Modules,Goal,Dp,Cs,Succ,[Var|Uvars]).

 prove_body(Modules,cv(sigma,(A -> o) -> o,c)^B,Dp,Cs,Succ,yes,Uvars) :-
	  !, new_para_var(Uvars,A,Var), lnorm(B ^ Var,Goal),
	  prove_goal(Modules,Goal,Dp,Cs,Succ,Uvars).

 prove_body(Modules,cv(=>,A->o->o,c)^cv(ModName,A,c)^Goal,Dp,Cs,Succ,yes,Uvars) :-
	  !, prove_goal([ModName|Modules],Goal,Dp,Cs,Succ,Uvars).

 prove_body(Modules,cv(clean_ff,o,c),Dp,Cs,Succ,yes,Uvars) :-
	  !, clean_ff_pairs(Dp,Succ,Cs).

 prove_body(Modules,Goal,Dp,Cs,Succ,yes,Uvars) :-
	  special_goal(Modules,Uvars,Goal,Extradps), 
	  (Extradps = [] , !, Succ = Dp, Cs=[] ;
	  append(Extradps,Dp,Dp1), unify(Dp1,Succ,Cs)).

 prove_body(Modules,Goal,Dp,Cs,Succ,yes,Uvars) :-
	  prove_atomic_goal(Modules,Goal,Dp,Cs,Succ,Uvars).



% Trying to solve an atomic goal. There are two cases. Either the goal has
% a variable head, or it has a constant head. In the former case the solution
% is obtained via a trivial substitution. In the latter case, a clause is 
% tried; here failure in a particular clause due to an attempt to backtrack 
% over a cut needs to result in a failure in trying to solve the atomic goal.

  prove_atomic_goal(Modules,Vargoal,Dp,Cs,Succ,Uvars) :-
        term_head(Vargoal,cv(Name,Ty,v)), !, target_and_args_type(Ty,_,Argtys),
        new_vars_from_types(Argtys,Bindvars),
        huet_normal(Substerm,hn(Bindvars,cv(true,o,c),[])),
        apply_subs_to_pairlist([[cv(Name,Ty,v),Substerm]],Dp,Newdp),
        unify(Newdp,Succ,Cs1), append([[cv(Name,Ty,v),Substerm]],Cs1,Cs).

  prove_atomic_goal(Modules,Goal,Dp,Cs,Succ,Uvars) :-
        (hptrace(goal,goal,'Entering goal',[Modules,Uvars,Goal]);
         hptrace(goal,goal,'Leaving goal', [Modules,Uvars,Goal]),!,fail),
        clause_instan(Uvars,Goal,Clause,Modules,NewModules),
        hptrace(clause,term,'Current clause',Clause),
        clause_head_and_body(Clause,Head,Body),
	abst_uvars(Uvars,Head,AbstHead),
	abst_uvars(Uvars,Goal,AbstGoal),
        unify([[AbstHead,AbstGoal]|Dp],Succ1,Cs1), 
	apply_subs_to_goal(Cs1,Uvars,Body,NewUvars,Newgoal),
        prove_body(NewModules,Newgoal,Succ1,Cs2,Succ,Backtrk,NewUvars),
        (Backtrk=yes; Backtrk=no,!,fail), append(Cs1,Cs2,Cs).

