
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                         %
%                           FILE  SCANNER.PRO                             %
%                         ---------------------                           %
%                                                                         %
%    This file contains a set of clauses that implement a scanner used    %
%   principally to read text (logical) line by line from a file. The main %
%   predicates in this regard are read_line(Line) and read_in(Line,Eoln). %
%   This code is adapted from Clocksin & Mellish and from Dale's imp:lex. %
%   The major difference from C&M is that an entire file has to be read.  %
%   The differences from Dale's code are that (1) the entire file is not  %
%   read in one shot, but rather text is read line by line, and (2) no    %
%   end-of-file marker is used but rather this is discovered by detecting %
%   an error condition. A lookahead character is used while determining   %
%   tokens. Consequently, the first character of the next line is read in %
%   the course of reading the current line. In order to make line by line %
%   processing possible, a trace of this character must be left, and this %
%   is done via an assertion of the form look_ahead(Char). The important  %
%   thing to note is that prior to reading a file, this assertion needs   %
%   to be initialised (eg look_ahead(32) - 32 is ASCII for blank - could  %
%   be used. The second point to note is that the detection of an end-    %
%   of-file condition depends crucially on the attempt to read beyond the %
%   end of a file being noted merely as a failure, and not being reported %
%   as an error. In other words, before a file is opened to be read, the  %
%   nofileerrors option must be turned on in CProlog. Finally, if an end- %
%   of-file is encountered then this results in an empty line being       %
%   returned by read_line. Any external routines that use this scanner    %
%   should check for this in order to determine whether the file is read  %
%   fully.                                                                %
%                                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




%          initialising some `constant' definitions

% the ASCIIs for characters that have a special role
    blank(Char) :- name(' ',[Char]).         % a space
    blank(Char) :- name('	',[Char]).   % a tab
    comment(Char) :- name('%',[Char]).       % start of a comment
    quote(Char) :- name('''',[Char]).        % start of a string
    quote(34).                               % Dale used this for strings. Why?
    eof(-1).                                 % needed at Duke
    eof(26).
    eoln(10).

    semicolon(';').

% Flush the rest of the (physical) line in the input text
    skip_line :- skip(10).



% Get the (ASCII for the) next char in input - it is a space if end-of-file.
    get_next_char(Char) :- (get0(Char), ! ; blank(Char)).




%  Get the next char in the input, skipping over spaces, comments, 
%  and special characters; Char is the lookahead
 get_next_nonblank(Char,NChar) :-
   ((blank(Char),! ; comment(Char),!,skip_line ; eoln(Char),! ; eof(Char),!),
         get0(Char1), get_next_nonblank(Char1,NChar) ;
    NChar = Char).





%  Get the first character of the current line
  get_lookahead(Char) :- (retract(look_ahead(Char)), ! ; blank(Char)).

%  Store the first character of the next line
  put_lookahead(Char) :- assert(look_ahead(Char)).




%  Read the next (logical) line in the input and return it as a list
%  of tokens.
     read_line(Line) :- read_in(Line,'.'), !.



%  Read the sequence of tokens upto the first one equal to EndWord and return
%  it as a list. If an end-of-file is encountered before EndWord then return
%  an empty list; in this case if the sequence is nonempty then flag an error.
  read_in(Words,EndWord) :-
      get_lookahead(Char), read_word(Char,Word,NextChar), !,
      (rest_sent(Word,NextChar,FinalChar,RestWords,EndWord), !,
                 Words = [Word|RestWords], put_lookahead(FinalChar) ;
       read_error(eof,EndWord), Words = []).

 read_in([],_).



%  Read in the next word from the input. Fail if end-of-file is encountered
%  before the start of a word. Char and NextChar are the lookaheads on entry
%  and on exit respectively
  read_word(Char,Word,NextChar) :- 
     get_next_nonblank(Char,TChar),
     (read_word1(TChar,Word,NextChar),! ; 
      read_error(badchar,TChar), blank(NChar), read_word(NChar,Word,NextChar)).


 read_word1(Char,Word,NextChar) :-
     single_character(Char), !, name(Word,[Char]), get_next_char(NextChar).

 read_word1(Char,Word,NextChar) :-
     in_word(Char), !, get_next_char(TChar),rest_word(TChar,Cs,NextChar),
     name(Word,[Char|Cs]).

 read_word1(Char,Word,NextChar) :-
     quote(Char), !, get_next_char(TChar),
     rest_string(TChar,Str,NextChar), name(Word,[Char|Str]).



% Get the remaining characters in a legal word
 rest_word(Char,[Char|Cs],NextChar) :-
     in_word(Char), !, get_next_char(TChar), rest_word(TChar,Cs,NextChar).

 rest_word(Char,[],Char).


% Get the remaining characters in a legal string.
 rest_string(Char,String,NextChar) :-
     quote(Char),!,get_next_char(NChr1),
     (quote(NChr1), 
        !, get_next_char(NChr2), rest_string(NChr2,RString,NextChar),
        String = [Char|RString] ;
      NextChar = NChr1, String = [Char]) .

 rest_string(Char,[Char|String],NextChar) :- !,
     get_next_char(Chr), rest_string(Chr,String,NextChar).


%  Get the remaining tokens upto EndWord; fail if an end-of-file is encountered
%  before an EndWord. Char and FinalChar are the lookahead characters before
%  entry and after exit respectively
  rest_sent(Word,Char,Char,[],Word) :- !.
 
  rest_sent(Word,Char,FinalChar,[NWord|Ws],EndWord) :-
      read_word(Char,NWord,NChar), rest_sent(NWord,NChar,FinalChar,Ws,EndWord).





% This is a list of characters that are legal in a word
in_word(33).           %  !  
in_word(35).           %  #
in_word(38).           %  &  
in_word(45).           %  -
in_word(47).           %  /  
in_word(58).           %  :
in_word(60).           %  <  
in_word(61).           %  =  
in_word(62).           %  >  
in_word(63).           %  ?
in_word(64).           %  @
in_word(95).           %  _
in_word(C) :- C>96, C<123.  % a,b,...,z
in_word(C) :- C>64, C<91.   % A,B,...,Z
in_word(C) :- C>47, C<58.   % 0,1,...,9





% These are a list of characters that form words by themselves
single_character(36).  %  $  
single_character(40).  %  (  
single_character(41).  %  )  
single_character(42).  %  *  
single_character(43).  %  +  
single_character(44).  %  ,  
single_character(46).  %  .  
%single_character(58).  %  :
single_character(59).  %  ;  
single_character(91).  %  [
single_character(92).  %  \  
single_character(93).  %  ]
single_character(94).  %  ^
single_character(96).  %  `
single_character(123). %  {  
single_character(124). %  |
single_character(125). %  }  
single_character(126). %  ~  



% This is sometimes useful for interacting with the user
   get_response(Word) :- blank(Char),!, read_word(Char,Word,_).


% This is useful in removing the double quotes from around a string
   remove_quotes(Str,Sans) :- 
     name(Str,[Quote|Rest]), reverse_cdr(Rest,Sanslist), name(Sans,Sanslist).

